@@ -29,7 +29,9 @@ from dagster_k8s.launcher import K8sRunLauncher
 from .client import DagsterKubernetesClient
 from .container_context import K8sContainerContext
 from .job import (
+    USER_DEFINED_K8S_CONFIG_SCHEMA,
     DagsterK8sJobConfig,
+    UserDefinedDagsterK8sConfig,
     construct_dagster_k8s_job,
     get_k8s_job_name,
     get_user_defined_k8s_config,
@@ -67,6 +69,11 @@ _K8S_EXECUTOR_CONFIG_SCHEMA = merge_dicts(
             ),
         ),
         "tag_concurrency_limits": get_tag_concurrency_limits_config(),
+        "step_k8s_config": Field(
+            USER_DEFINED_K8S_CONFIG_SCHEMA,
+            is_required=False,
+            description="Raw Kubernetes configuration for each step launched by the executor.",
+        ),
     },
 )
 
@@ -133,6 +140,10 @@ def k8s_job_executor(init_context: InitExecutorContext) -> Executor:
         namespace=exc_cfg.get("job_namespace"),  # type: ignore
         resources=exc_cfg.get("resources"),  # type: ignore
         scheduler_name=exc_cfg.get("scheduler_name"),  # type: ignore
+        security_context=exc_cfg.get("security_context"),  # type: ignore
+        # step_k8s_config feeds into the run_k8s_config field because it is merged
+        # with any configuration for the run that was set on the run launcher or code location
+        run_k8s_config=UserDefinedDagsterK8sConfig.from_dict(exc_cfg.get("step_k8s_config", {})),
     )
 
     if "load_incluster_config" in exc_cfg:
@@ -215,7 +226,7 @@ class K8sStepHandler(StepHandler):
         user_defined_k8s_config = get_user_defined_k8s_config(
             step_handler_context.step_tags[step_key]
         )
-        return context.merge(K8sContainerContext(run_k8s_config=user_defined_k8s_config.to_dict()))
+        return context.merge(K8sContainerContext(run_k8s_config=user_defined_k8s_config))
 
     def _get_k8s_step_job_name(self, step_handler_context: StepHandlerContext):
         step_key = self._get_step_key(step_handler_context)
@@ -256,23 +267,29 @@ class K8sStepHandler(StepHandler):
         if not job_config.job_image:
             raise Exception("No image included in either executor config or the job")
 
+        run = step_handler_context.dagster_run
+        labels = {
+            "dagster/job": run.job_name,
+            "dagster/op": step_key,
+            "dagster/run-id": step_handler_context.execute_step_args.run_id,
+        }
+        if run.external_job_origin:
+            labels[
+                "dagster/code-location"
+            ] = run.external_job_origin.external_repository_origin.code_location_origin.location_name
         job = construct_dagster_k8s_job(
             job_config=job_config,
             args=args,
             job_name=job_name,
             pod_name=pod_name,
             component="step_worker",
-            user_defined_k8s_config=container_context.get_run_user_defined_k8s_config(),
-            labels={
-                "dagster/job": step_handler_context.dagster_run.job_name,
-                "dagster/op": step_key,
-                "dagster/run-id": step_handler_context.execute_step_args.run_id,
-            },
+            user_defined_k8s_config=container_context.run_k8s_config,
+            labels=labels,
             env_vars=[
                 *step_handler_context.execute_step_args.get_command_env(),
                 {
                     "name": "DAGSTER_RUN_JOB_NAME",
-                    "value": step_handler_context.dagster_run.job_name,
+                    "value": run.job_name,
                 },
                 {"name": "DAGSTER_RUN_STEP_KEY", "value": step_key},
                 *container_context.env,
