@@ -59,6 +59,7 @@ class K8sRunLauncher(RunLauncher, ConfigurableClass):
         env_secrets=None,
         env_vars=None,
         k8s_client_batch_api=None,
+        k8s_client_core_api=None,
         volume_mounts=None,
         volumes=None,
         labels=None,
@@ -67,6 +68,8 @@ class K8sRunLauncher(RunLauncher, ConfigurableClass):
         scheduler_name=None,
         security_context=None,
         run_k8s_config=None,
+        only_allow_user_defined_k8s_config_fields=None,
+        only_allow_user_defined_env_vars=None,
     ):
         self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
         self.job_namespace = check.str_param(job_namespace, "job_namespace")
@@ -84,7 +87,8 @@ class K8sRunLauncher(RunLauncher, ConfigurableClass):
             kubernetes.config.load_kube_config(kubeconfig_file)
 
         self._api_client = DagsterKubernetesClient.production_client(
-            batch_api_override=k8s_client_batch_api
+            core_api_override=k8s_client_core_api,
+            batch_api_override=k8s_client_batch_api,
         )
 
         self._job_config = None
@@ -118,6 +122,9 @@ class K8sRunLauncher(RunLauncher, ConfigurableClass):
         self._scheduler_name = check.opt_str_param(scheduler_name, "scheduler_name")
         self._security_context = check.opt_dict_param(security_context, "security_context")
         self._run_k8s_config = check.opt_dict_param(run_k8s_config, "run_k8s_config")
+
+        self._only_allow_user_defined_k8s_config_fields = only_allow_user_defined_k8s_config_fields
+        self._only_allow_user_defined_env_vars = only_allow_user_defined_env_vars
         super().__init__()
 
     @property
@@ -180,6 +187,14 @@ class K8sRunLauncher(RunLauncher, ConfigurableClass):
     def fail_pod_on_run_failure(self) -> Optional[bool]:
         return self._fail_pod_on_run_failure
 
+    @property
+    def only_allow_user_defined_k8s_config_fields(self) -> Optional[Mapping[str, Any]]:
+        return self._only_allow_user_defined_k8s_config_fields
+
+    @property
+    def only_allow_user_defined_env_vars(self) -> Optional[Sequence[str]]:
+        return self._only_allow_user_defined_env_vars
+
     @classmethod
     def config_type(cls):
         """Include all arguments required for DagsterK8sJobConfig along with additional arguments
@@ -206,17 +221,27 @@ class K8sRunLauncher(RunLauncher, ConfigurableClass):
         pod_name = job_name
 
         job_origin = check.not_none(run.job_code_origin)
-        user_defined_k8s_config = container_context.get_run_user_defined_k8s_config()
+        user_defined_k8s_config = container_context.run_k8s_config
         repository_origin = job_origin.repository_origin
 
         job_config = container_context.get_k8s_job_config(
             job_image=repository_origin.container_image, run_launcher=self
         )
+        job_image = job_config.job_image
+        if job_image:  # expected to be set
+            self._instance.add_run_tags(
+                run.run_id,
+                {DOCKER_IMAGE_TAG: job_image},
+            )
 
-        self._instance.add_run_tags(
-            run.run_id,
-            {DOCKER_IMAGE_TAG: job_config.job_image},
-        )
+        labels = {
+            "dagster/job": job_origin.job_name,
+            "dagster/run-id": run.run_id,
+        }
+        if run.external_job_origin:
+            labels[
+                "dagster/code-location"
+            ] = run.external_job_origin.external_repository_origin.code_location_origin.location_name
 
         job = construct_dagster_k8s_job(
             job_config=job_config,
@@ -225,10 +250,7 @@ class K8sRunLauncher(RunLauncher, ConfigurableClass):
             pod_name=pod_name,
             component="run_worker",
             user_defined_k8s_config=user_defined_k8s_config,
-            labels={
-                "dagster/job": job_origin.job_name,
-                "dagster/run-id": run.run_id,
-            },
+            labels=labels,
             env_vars=[
                 {
                     "name": "DAGSTER_RUN_JOB_NAME",
@@ -346,7 +368,9 @@ class K8sRunLauncher(RunLauncher, ConfigurableClass):
     def supports_run_worker_crash_recovery(self):
         return True
 
-    def get_run_worker_debug_info(self, run: DagsterRun) -> Optional[str]:
+    def get_run_worker_debug_info(
+        self, run: DagsterRun, include_container_logs: Optional[bool] = True
+    ) -> Optional[str]:
         container_context = self.get_container_context_for_run(run)
         if self.supports_run_worker_crash_recovery:
             resume_attempt_number = self._instance.count_resume_run_attempts(run.run_id)
@@ -355,23 +379,19 @@ class K8sRunLauncher(RunLauncher, ConfigurableClass):
 
         job_name = get_job_name_from_run_id(run.run_id, resume_attempt_number=resume_attempt_number)
         namespace = container_context.namespace
-        user_defined_k8s_config = container_context.get_run_user_defined_k8s_config()
-        container_name = user_defined_k8s_config.container_config.get("name", "dagster")
         pod_names = self._api_client.get_pod_names_in_job(job_name, namespace=namespace)
         full_msg = ""
         try:
             pod_debug_info = [
                 self._api_client.get_pod_debug_info(
-                    pod_name, namespace, container_name=container_name
+                    pod_name, namespace, include_container_logs=include_container_logs
                 )
                 for pod_name in pod_names
             ]
             full_msg = "\n".join(pod_debug_info)
         except Exception:
             logging.exception(
-                "Error trying to get debug information for failed k8s job {job_name}".format(
-                    job_name=job_name
-                )
+                f"Error trying to get debug information for failed k8s job {job_name}"
             )
         if pod_names:
             full_msg = (
@@ -382,9 +402,12 @@ class K8sRunLauncher(RunLauncher, ConfigurableClass):
             )
 
         else:
+            job_debug_info = self._api_client.get_job_debug_info(job_name, namespace=namespace)
             full_msg = (
                 full_msg
-                + "\nFor more information about the failure, try running `kubectl describe job"
+                + "\n\n"
+                + job_debug_info
+                + "\n\nFor more information about the failure, try running `kubectl describe job"
                 f" {job_name}` in your cluster."
             )
 
